import std/[cookies, json, options, sequtils, strtabs, strutils, tables, times]
import mummy, uuids, webby
import std/httpclient except HttpHeaders

import jwt_full, jwt_full/encoding

import ./[apiutils,jsonutils]

const SUPPORTED_SIGNATURE_ALGORITHMS = @[ HS256, RS256 ]


type
  AuthError* = object of CatchableError
    additionalInfo*: Option[TableRef[string, JsonNode]]

  ApiAuthContext* = ref object
    appDomain*: string            ## Application domain for session cookies
    cookiePrefix*: string         ## Prefix for the user and session cookies
    validAudiences*: seq[string]  ## Expected audience values for for `aud` JWT check
    issuer*: string               ## The JWT issuer for tokens created by this API
    trustedIssuers*: seq[string]  ## The list of trusted OAuth 2.0 issuers

    signingKid*: string
      ## The key id to use for signing JWT tokens created by this API. The JWK
      ## representing this signing key with the private or secret key value
      ## must be provided either when the ApiAuthContext is initialized (see
      ## `initApiAuthContext` or via `addSigningKeys`

    issuerKeys: TableRef[string, JwkSet]


proc failAuth*[T](
    reason: string,
    additionalInfo: TableRef[string, T],
    parentException: ref Exception = nil) =
  ## Syntactic sugar to raise an AuthError. Reason will be the exception
  ## message and should be considered an internal message.
  let err = newException(AuthError, reason, parentException)
  err.additionalInfo = some(newTable[string, JsonNode]())
  for key, val in additionalInfo.pairs:
    err.additionalInfo.get[key] = %val
  raise err


proc failAuth*(reason: string, parentException: ref Exception = nil) =
  ## Syntactic sugar to raise an AuthError. Reason will be the exception
  ## message and should be considered an internal message.
  raise newException(AuthError, reason, parentException)


proc validateSigningKey(k: JWK): void =
  if k.alg.isNone: failAuth "JWK is missing 'alg'"
  if k.kid.isNone: failAuth "JWK is missing 'kid'"


proc initApiAuthContext*(
    appDomain: string,
    cookiePrefix: string,
    validAudiences: seq[string],
    issuer: string,
    trustedIssuers: seq[string],
    signingKid: string,
    signingKeys: JwkSet): ApiAuthContext =

  for k in signingKeys: validateSigningKey(k)

  result = ApiAuthContext(
    appDomain: appDomain,
    cookiePrefix: cookiePrefix,
    validAudiences: validAudiences,
    issuer: issuer,
    trustedIssuers: trustedIssuers,
    signingKid: signingKid,
    issuerKeys: newTable[string, JwkSet]([(issuer, signingKeys)]))


proc fetchJWKs(openIdConfigUrl: string): JwkSet {.gcsafe.} =
  ## Fetch signing keys for an OAuth issuer. `openIdConfigUrl` is expected to
  ## be a well-known URL (ISSUER_BASE/.well-known/openid-configuration)
  var jwksKeysURI: string
  try:
    let http = newHttpClient()

    # Inspect the OAuth metadata via the well-known address.
    let metadata = parseJson(http.getContent(openIdConfigUrl))

    # Fetch the keys from the jwk_keys URI.
    jwksKeysURI = metadata.getOrFail("jwks_uri").getStr
    let jwksKeys = parseJson(http.getContent(jwksKeysURI))

    # Parse and load the keys provided.
    return initJwkSet(jwksKeys)

  except Exception:
    #failAuth "unable to fetch isser signing keys"
    failAuth(
      reason = "unable to fetch isser signing keys",
      additionalInfo = newTable[string, string]({
        "openIdConfigUrl": openIdConfigUrl,
        "jwksKeysURI": jwksKeysURI }),
      parentException = getCurrentException())


proc addSigningKeys*(ctx: ApiAuthContext, issuer: string, keySet: JwkSet): void =
  ## Manually add a set of signing keys associated with a given issuer.
  try:
    for k in keySet: validateSigningKey(k)
    if ctx.issuerKeys.isNil: ctx.issuerKeys = newTable[string, JwkSet]()
    ctx.issuerKeys[issuer] = keySet
  except:
    raise getCurrentException()


proc findSigningKey*(ctx: ApiAuthContext, jwt: JWT, allowFetch = true): JWK {.gcsafe.} =
  ## Lookup the signing key for a given JWT. This method assumes that you trust
  ## the issuer named in the JWT.
  ##
  ## If our ApiAuthContext does not contain any keys for the issuer named in
  ## the JWT, or if that set of keys does not contain the key id referenced in
  ## the JWT, we will attempt to fetch the signing keys based on the
  ## [OpenID Connect standard discovery mechanism](https://openid.net/specs/openid-connect-discovery-1_0.html)

  try:
    if jwt.claims.iss.isNone: failAuth "JWT is missing 'iss' claim."
    if jwt.header.kid.isNone: failAuth "JWT is missing 'kid' header."

    if ctx.issuerKeys.isNil: ctx.issuerKeys = newTable[string, JwkSet]()

    let jwtIssuer = jwt.claims.iss.get

    if ctx.issuerKeys.hasKey(jwtIssuer):
      let foundKeys = ctx.issuerKeys[jwtIssuer]
        .filterIt(it.kid.isSome and it.kid.get == jwt.header.kid.get)

      if foundKeys.len == 1: return foundKeys[0]

    # If we didn't have keys for that issuer, or if we couldn't find the given
    # kid, we need to refresh our cache of keys.
    if allowFetch:
      ctx.issuerKeys[jwtIssuer] =
        fetchJWKs(jwtIssuer & "/.well-known/openid-configuration")
      return ctx.findSigningKey(jwt, false)

    failAuth(
      reason = "unable to find JWT signing key",
      additionalInfo = newTable[string, string]({
        "jwtIssuer": jwtIssuer,
        "jwtKid": jwt.header.kid.get}))

  except:
    failAuth("unable to find JWT signing key", getCurrentException())


proc validateJWT*(ctx: ApiAuthContext, jwt: JWT) =
  ## Given a JWT, validate that it is a well-formed JWT, validate the issuer's
  ## signature on the token, and validate all the claims that it preesnts.
  try:
    if jwt.claims.iss.isNone: failAuth "JWT is missing 'iss' claim."
    let jwtIssuer = jwt.claims.iss.get

    if not ctx.trustedIssuers.contains(jwtIssuer):
      failAuth(
        reason = "We don't trust the JWT's issuer.",
        additionalInfo = newTable[string, JsonNode]({
          "issuer": %jwtIssuer,
          "trustedIssuers": %ctx.trustedIssuers}))

    if jwt.header.alg.isNone: failAuth "JWT is missing 'alg' header property."
    if jwt.claims.aud.isNone: failAuth "JWT is missing 'aud' claim."
    if jwt.claims.iss.isNone: failAuth "JWT is missing 'iss' claim."
    if jwt.claims.sub.isNone: failAuth "JWT is missing 'sub' claim."
    if jwt.claims.exp.isNone: failAuth "JWT is missing or invalid 'exp' claim."

    if jwt.claims["aud"].get.kind == JString:
      # If the token is for a single audience, check that it is for us.
      if not ctx.validAudiences.contains(jwt.claims.aud.get):
        failAuth "JWT is not for us (invalid audience)."
    elif jwt.claims["aud"].get.kind == JArray:
      # If the token is for multiple audiences, check that at least one is for us.
      let auds = jwt.claims["aud"].get.getElems
      if not auds.anyIt(ctx.validAudiences.contains(it.getStr)):
        failAuth "JWT is not for us (invalid audience)."

    let signingAlgorithm = jwt.header.alg.get

    if not SUPPORTED_SIGNATURE_ALGORITHMS.contains(signingAlgorithm):
      failAuth "unacceptable signature algorithm: " & $signingAlgorithm

    jwt.validate(
      sigAlg = signingAlgorithm,
      key = ctx.findSigningKey(jwt),
      validateTimeClaims = true)

  except:
    failAuth(getCurrentExceptionMsg(), getCurrentException())


proc extractValidJwt*(ctx: ApiAuthContext, req: Request, validateCsrf = true): JWT =
  ## Extracts a valid JWT representing the user's authentication and
  ## authorization details, if present. If there are no valid credentials an
  ## exception is raised.
  ##
  ## We support two authentication flows:
  ##
  ## - Strict API via a JWT Bearer token in the Authorization header. This is
  ##   intended for API consumers (not a browser-based web-app). In this case,
  ##   the token is validated directly.
  ##
  ## - Split JWT via two cookies:
  ##
  ##   - `${cookiePrefix}-user`: Contains the JWT header and payload, but not the
  ##      signature. This cookie should be set Secure. The JWT payload should
  ##      have a defined expiration date (matching the Max-Age of the cookie)
  ##      and a CSRF token. This cookie is accessible by the web application.
  ##
  ##   - `${cookiePrefix}-session`: Contains the JWT signature. This cookie is
  ##      set Secure and HttpOnly. This serves as the session token (when the
  ##      user closes the browser this gets unset).
  ##
  ##   In the split-cookie mode we also check that the `csrfToken` claim in the
  ##   JWT payload matches the CSRF value passed via the `X-CSRF-TOKEN` header.
  ##   This CSRF check can be disabled by setting `validateCsrf` to `false`.
  ##   This option is proivded to support occasional use-cases where you want to
  ##   be able to serve a request using cookie auth when the client can't set
  ##   custom headers (e.g. a simple link from an <a> tag). Obviously, this is a
  ##   security risk and should only be used with caution with a full
  ##   understanding of the risk.

  try:
    if req.headers.contains("Authorization"):
      # Using a Bearer token.
      result = toJWT(req.headers["Authorization"][7..^1])

    elif req.headers.contains("Cookie"):
      # Using a user/session cookie pair
      let userCookieName = ctx.cookiePrefix & "-user"
      let sessionCookieName = ctx.cookiePrefix & "-session"

      let cookies = parseCookies(req.headers["Cookie"])
      if not cookies.contains(userCookieName):
        failAuth "missing cookie '$#'" % userCookieName
      if not cookies.contains(sessionCookieName):
        failAuth "missing cookie '$#'" % sessionCookieName

      let userVal = cookies[userCookieName]
      let sessionVal = cookies[sessionCookieName]
      result = toJWT(userVal & "." & sessionVal)

      # Because this is a web session, check that the CSRF is present and
      # matches.
      if validateCsrf:
        if not req.headers.contains("X-CSRF-TOKEN") or
           not result.claims["csrfToken"].isSome:
          failAuth "missing CSRF token"

        if req.headers["X-CSRF-TOKEN"] != result.claims["csrfToken"].get.getStr(""):
          failAuth(
            reason = "invalid CSRF token",
            additionalInfo = newTable[string, string]({
              "header": req.headers["X-CSRF-TOKEN"],
              "jwt": result.claims["csrfToken"].get.getStr("")}))

    else: failAuth "no auth token, no Authorization or Cookie headers"

    ctx.validateJWT(result)
  except:
    failAuth(getCurrentExceptionMsg(), getCurrentException())


proc createSessionCookies*(ctx: ApiAuthContext, jwt: JWT): HttpHeaders =

  # Split the token to get the user and session cookie values.
  let strToken = $jwt
  let splitToken = strToken.rsplit('.', 1)

  # User cookie (accessible by the application)
  let userCookie = setCookie(
    key = ctx.cookiePrefix & "-user",
    value = splitToken[0],
    domain = ctx.appDomain,
    expires = jwt.claims.exp.get.utc,
    httpOnly = false,
    path = "/",
    sameSite = SameSite.Strict,
    secure = true)

  # Session cookie (used by the API)
  let sessionCookie = setCookie(
    key = ctx.cookiePrefix & "-session",
    value = splitToken[1],
    domain = ctx.appDomain,
    httpOnly = true,
    path = "/",
    sameSite = SameSite.Strict,
    secure = true)

  for c in [userCookie, sessionCookie]:
    let parts = c.split(": ")
    result &= [(parts[0], parts[1])]


proc createSignedJWT*(ctx: ApiAuthContext, claims: JsonNode, kid: string): JWT =
  ## Given a set of claims, create a JWT using the given key for our issuer
  ## (as defined in the ApiAuthContext). This is an opinionated method that
  ## chooses the signing algorithm,

  let foundKeys = ctx.issuerKeys[ctx.issuer]
    .filterIt(it.kid.isSome and it.kid.get == kid)

  if foundKeys.len != 1:
    failAuth "cannot create signed JWT, unable to find key for kid " & kid

  let sigKey = foundKeys[0]

  result = createSignedJwt(
    initJoseHeader(%*{
      "alg": sigKey.alg.get,
      "typ": "JWT",
      "kid": sigKey.kid.get }),
    initJwtClaims(claims),
    sigKey)


proc newApiAccessToken*(ctx: ApiAuthContext, sub: string, duration = 1.hours): JWT =
  ## Create a new JWT for API access.
  result = ctx.createSignedJWT(
    %*{
      "sid": $genUUID(),
      "sub": sub,
      "iss": ctx.issuer,
      "iat": now().utc.toTime.toUnix.int,
      "aud": ctx.issuer,
      "exp": (now() + duration).utc.toTime.toUnix.int },
    ctx.signingKid)
